// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Authors: See the file AUTHORS for full details.
//
//  Copyright (C) 2002-2004 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
// $Id: simplify.cc,v 1.4 2008-08-21 15:34:11 carzanig Exp $
//
#include <cassert>
#include <iostream>
#include <list>
#include <memory>
#include <stdexcept>
#include "siena/simplify.h"

using namespace siena;
using namespace std;

#include "siena_simplify_impl.h"

using namespace siena_simplify_impl;

bool siena::conflicting( const constraint& c1, const constraint& c2 )
{
  // constraints never conflict with themselves or with
  // other constraints with different names.
  if( &c1 == &c2 || !(c2.name() == c1.name()) )
    {
      return false;
    }
  // constraints always conflict if their types are different
  if( c2.type() != c1.type() )
    {
      return true;
    }
  switch( c1.op() )
    {
    case eq_id:
      switch( c2.op() )
	{
	case eq_id:
	  // x = a and x = b (b != a)
	  return neq( (value&)c2, (value&)c1 );
	case lt_id:
	  // x = a and x < b (b <= a)
	  return leq( (value&)c2, (value&)c1 );
	case gt_id:
	  // x = a and x > b (b >= a)
	  return geq( (value&)c2, (value&)c1 );
	case sf_id:
	  // x = a and x *= b !(a *= b)
	  return ! sf( c1.string_value(), c2.string_value() );
	case pf_id:
	  // x = a and x =* b !(a =* b)
	  return ! pf( c1.string_value(), c2.string_value() );
	case ss_id:
	  // x = a and x ** b !(a ** b)
	  return ! ss( c1.string_value(), c2.string_value() );
	case ne_id:
	  // x = a and x != b (b = a)
	  return ! neq( (value&)c1, (value&)c2 );
	default:
	  return false;
	}
    case lt_id:
      switch( c2.op() )
	{
	case gt_id:
	case eq_id:
	  // x < a and x = b (b >= a)
	  // x < a and x > b (b >= a)
	  return geq( (value&)c2, (value&)c1 );
	default:
	  return false;
	}
    case gt_id:
      switch( c2.op() )
	{
	case eq_id:
	case lt_id:
	  // x > a and x = b (b <= a)
	  // x > a and x < b (b <= a)
	  return leq( (value&)c2, (value&)c1 );
	default:
	  return false;
	}
    case sf_id:
      switch( c2.op() )
	{
	case eq_id:
	  // x *= a and x = b !(b *= a)
	  return ! sf( c2.string_value(), c1.string_value() );
	case sf_id:
	  // x *= a and x *= b !(a *= b) && !(b *= a)
	  return !sf( c1.string_value(), c2.string_value() ) &&
	    !sf( c2.string_value(), c1.string_value() );
	default:
	  return false;
	}
    case pf_id:
      switch( c2.op() )
	{
	case eq_id:
	  // x =* a and x = b !(a =* b)
	  return ! pf( c1.string_value(), c2.string_value() );
	case pf_id:
	  // x =* a and x =* b !(a =* b) && !(b =* a)
	  return !pf( c1.string_value(), c2.string_value() ) &&
	    !pf( c2.string_value(), c1.string_value() );
	default:
	  return false;
	}
    case ss_id:
      switch( c2.op() )
	{
	case eq_id:
	  // x ** a and x = b  !(b ** a)
	  return !ss( c2.string_value(), c1.string_value() );
	default:
	  return false;
	}
    case ne_id:
      switch( c2.op() )
	{
	case eq_id:
	  // x != a and x = b (a = b)
	  return ! neq( (value&)c1, (value&)c2 );
	default:
	  return false;
	}
    default:
      // want to make sure we have covered all cases
      assert( false );
    }
}

bool siena::covers( const predicate& p1, const predicate& p2 )
{
  /*
   * if each filter of p2 is covered by a filter in p1 then p1
   * covers p2.
   */
  auto_ptr<predicate::iterator> p2itr( p2.first() );
  if( p2itr.get() == NULL )
    {
      // an empty predicate is always covered even if p1 is empty too
      return true;
    }
  do
    {
      bool covered = false;
      auto_ptr<predicate::iterator> p1itr( p1.first() );
      if( p1itr.get() != NULL )
	{
	  do
	    {
	      if( covers( (*p1itr), (*p2itr) ) )
		{
		  covered = true;
		}
	    }
	  while( !covered && p1itr->next() );
	}
      if( !covered )
	{
	  return false;
	}
    }
  while( p2itr->next() );

  return true;
}

bool siena::covers( const filter& f1, const filter& f2 )
{

  // f1 covers f2 if each of its constraints covers
  // a constraint in f2
   
  auto_ptr<filter::iterator> f1itr( f1.first() );
  if( f1itr.get() == NULL )
    {
      // an empty filter covers nothing
      return false;
    }
  do
    {
      auto_ptr<filter::iterator> f2itr( f2.first() );
      if( f2itr.get() == NULL )
	{
	  // an empty filter can never be covered
	  return false;
	}
      bool found = false;
      do
	{
	  if( covers( (*f1itr), (*f2itr) ) )
	    {
	      found = true;
	    }
	}
      while( !found && f2itr->next() );
      if( !found )
	{
	  // we didn't find a constraint that
	  // is covered by a constraint in f1.
	  return false;
	}
    }
  while( f1itr->next() );

  // if here, every constraint in f1 covers
  // something in f2.
  return true;
}

bool siena::covers( const constraint& c1, const constraint& c2 )
{
  // if they are the same object or have different names,
  // always false
  if( &c1 == &c2 || !(c1.name() == c2.name()) )
    {
      return false;
    }

  // if c1's operator is any_id...
  if( c1.op() == any_id )
    {
       return c1.type() == anytype_id || c1.type() == c2.type();
    }

  // after eliminating any_id, if types are different,
  // always false
  if( c1.type() != c2.type() )
    {
      return false;
    }

  // if the two are equal, always true
  if( eq( c1, c2 ) )
    {
      return true;
    }

  // now do the special cases
  switch( c1.op() )
    {
    case lt_id:
      switch( c2.op() )
	{
	case lt_id:
	  // x < a covers x < b (b < a)
	case eq_id:
	  // x < a covers x = b (b < a)
	  return lt( (value&)c2, (value&)c1 );
	default:
	  return false;
	}
    case gt_id:
      switch( c2.op() )
	{
	case gt_id:
	  // x > a covers x > b (b > a)
	case eq_id:
	  // x > a covers x = b (b > a)
	  return gt( (value&)c2, (value&)c1 );
	default:
	  return false;
	}
    case ne_id:
      switch( c2.op() )
	{
	case lt_id:
	  // x != a covers x < b (b <= a)
	  return leq( (value&)c2, (value&)c1 );
	case gt_id:
	  // x != a covers x > b (b >= a)
	  return geq( (value&)c2, (value&)c1 );
	case eq_id:
	  // x > a covers x = b (b != a)
	  return neq( (value&)c2, (value&)c1 );
	default:
	  return false;
	}
    case sf_id:
      switch( c2.op() )
	{
	case sf_id:
	case eq_id:
	  // x *= a covers x = b if b *= a
	  return sf( c2.string_value(), c1.string_value() );
	case ss_id:
	  // x *= a covers x ** b if a ** b
	  return ss( c1.string_value(), c2.string_value() );
	default:
	  return false;
	}
    case pf_id:
      switch( c2.op() )
	{
	case pf_id:
	case eq_id:
	  // x =* a covers x = b if b =* a
	  return pf( c2.string_value(), c1.string_value() );
	case ss_id:
	  // x =* a covers x ** b if a ** b
	  return ss( c1.string_value(), c2.string_value() );
	default:
	  return false;
	}
    case ss_id:
      switch( c2.op() )
	{
	case ss_id:
	case pf_id:
	case sf_id:
	case eq_id:
	  // x ** a covers x = b if a ** b
	  return ss( c2.string_value(), c1.string_value() );
	default:
	  return false;
	}
    default:
      return false;
    }
}

bool siena_simplify_impl::geq( const value& v1, const value& v2 )
{
  if( v1.type() != v2.type() )
    {
      return false;
    }
  switch( v1.type() )
    {
    case string_id:
      return geq( v1.string_value(), v2.string_value() );
    case int_id:
      return v1.int_value() >= v2.int_value();
    case double_id:
      return v1.double_value() >= v2.double_value();
    case bool_id:
      return v1.bool_value() >= v2.bool_value();
    default:
      cerr << __FILE__ << ":" << __LINE__ << " UNKNOWN TYPE" << endl;
      return false;
    }
}

bool siena_simplify_impl::gt( const value& v1, const value& v2 )
{
  if( v1.type() != v2.type() )
    {
      return false;
    }
  switch( v1.type() )
    {
    case string_id:
      return gt( v1.string_value(), v2.string_value() );
    case int_id:
      return v1.int_value() > v2.int_value();
    case double_id:
      return v1.double_value() > v2.double_value();
    case bool_id:
      return v1.bool_value() > v2.bool_value();
    default:
      assert( false );
      return false;
    }
}

bool siena_simplify_impl::neq( const value& v1, const value& v2 )
{
  if( v1.type() != v2.type() )
    {
      return true;
    }
  switch( v1.type() )
    {
    case string_id:
      return !(v1.string_value() == v2.string_value());
    case int_id:
      return v1.int_value() != v2.int_value();
    case double_id:
      return v1.double_value() != v2.double_value();
    case bool_id:
      return v1.bool_value() != v2.bool_value();
    default:
      assert( false );
    }
}

bool siena_simplify_impl::leq( const value& v1, const value& v2 )
{
  if( v1.type() != v2.type() )
    {
      return false;
    }
  switch( v1.type() )
    {
    case string_id:
      return leq( v1.string_value(), v2.string_value() );
    case int_id:
      return v1.int_value() <= v2.int_value();
    case double_id:
      return v1.double_value() <= v2.double_value();
    case bool_id:
      return v1.bool_value() <= v2.bool_value();
    default:
      assert( false );
    }
}

bool siena_simplify_impl::lt( const value& v1, const value& v2 )
{
  if( v1.type() != v2.type() )
    {
      return false;
    }
  switch( v1.type() )
    {
    case string_id:
      return v1.string_value() < v2.string_value();
    case int_id:
      return v1.int_value() < v2.int_value();
    case double_id:
      return v1.double_value() < v2.double_value();
    case bool_id:
      return v1.bool_value() < v2.bool_value();
    default:
      assert( false );
    }
}

bool siena_simplify_impl::leq( const string_t& s1, const string_t& s2 )
{
  const char * i = s1.begin;
  const char * j = s2.begin;
  while(i != s1.end) {
    if (j == s2.end) return false;
    if (*i < *j) return true;
    if (*i > *j) return false;
    ++i;
    ++j;
  }
  return true;
}

bool siena_simplify_impl::gt( const string_t& s1, const string_t& s2 )
{
  const char * i = s1.begin;
  const char * j = s2.begin;
  while(i != s1.end) {
    if (j == s2.end) return true;
    if (*i > *j) return true;
    if (*i < *j) return false;
    ++i;
    ++j;
  }
  return false;
}

bool siena_simplify_impl::geq( const string_t& s1, const string_t& s2 )
{
  const char * i = s1.begin;
  const char * j = s2.begin;
  while(i != s1.end) {
    if (j == s2.end) return true;
    if (*i > *j) return true;
    if (*i < *j) return false;
    ++i;
    ++j;
  }
  return j == s2.end;
}

bool siena_simplify_impl::eq( const constraint& c1, const constraint& c2 )
{
  if( c1.op() != c2.op() || !(c1.name() == c2.name()) || c1.type() != c2.type() )
    {
      return false;
    }
  switch( c1.type() )
    {
    case string_id:
      return c1.string_value() == c2.string_value();
    case int_id:
      return c1.int_value() == c2.int_value();
    case double_id:
      return c1.double_value() == c2.double_value();
    case bool_id:
      return c1.bool_value() == c2.bool_value();
    default:
      assert( false );
    }
}

bool siena_simplify_impl::pf( const string_t& s1, const string_t& s2 )
{
    return s1.has_prefix(s2);
}

bool siena_simplify_impl::sf( const string_t& s1, const string_t& s2 )
{
    return s1.has_suffix(s2);
}

bool siena_simplify_impl::ss( const string_t& s1, const string_t& s2 )
{
    return s1.has_substring(s2);
}

